
#include <stdio.h>
#include <stdint.h>
#include <math.h>
#include <stdlib.h>

#include <SDL.h>
#include <SDL_image.h>

#define LTSIZE 20000
#define RESOL 16
#define PROMOTE 64
#define BIGPROMOTE 1024

static int intsqrt(unsigned int a)
{
    unsigned int root = 0, rem = 0;
    int i;

    for(i = 0; i < 16; i++)
    {
        root <<= 1;
        rem = (rem << 2) + (a >> 30);
        a <<= 2;
        root++;
        if(root <= rem)
        {
            rem -= root;
            root++;
        }
        else
            root--;
    }

    return root >> 1;
}

int normal(SDL_Surface *src, SDL_Surface *dst, int cx, int cy, int R,
           int d, double phi, double theta, double gamma)
{
    uint32_t *spix, *dpix;
    struct { int x, y; } pt[LTSIZE], *tmp, *tmp2;
    int i, j, di, dj, dx, dy;
    int cosg, sing, cosp, sinp, cost, sint;
    int srcpitch, dstpitch;

    dx = dst->w / RESOL;
    dy = dst->h / RESOL;

    if((dx + 1) * (dy + 1) > LTSIZE)
    {
        fprintf(stderr, "%ix%i too big for %i\n", dx, dy, LTSIZE);
        return -1;
    }

    spix = (uint32_t *)src->pixels;
    srcpitch = src->pitch / 4;
    dpix = (uint32_t *)dst->pixels;
    dstpitch = dst->pitch / 4;

    /* Précalcul ; on pourrait mettre ça dans une LUT, mais à quoi bon ? */
    cosg = BIGPROMOTE * cos(gamma);
    sing = BIGPROMOTE * sin(gamma);
    cosp = BIGPROMOTE * cos(phi);
    sinp = BIGPROMOTE * sin(phi);
    cost = BIGPROMOTE * cos(theta);
    sint = BIGPROMOTE * sin(theta);

    /* 1: on remplit le lookup */
    for(tmp = pt, j = 0; j <= dy; j++)
    {
        for(i = 0; i <= dx; i++)
        {
            int x1, y1, x2, y2, x3, y3, x4, y4, x5, y5;
            int t1, t2, t3;

            x5 = i * RESOL - dst->w / 2;
            y5 = j * RESOL - dst->h / 2;

            x4 = (x5 * cosg + y5 * sing) / (BIGPROMOTE / PROMOTE);
            y4 = (y5 * cosg - x5 * sing) / (BIGPROMOTE / PROMOTE);

            t3 = (d * PROMOTE) * (d * PROMOTE) + x4 * x4 + y4 * y4;
            t3 = intsqrt(1 + t3);
            x3 = x4 * R * PROMOTE / t3;
            y3 = y4 * R * PROMOTE / t3;

            t1 = R * PROMOTE;
            t2 = t1 * t1 - x3 * x3 - y3 * y3;
            x2 = x3;
            y2 = (y3 * cosp + intsqrt(t2) * sinp) / BIGPROMOTE;

            x1 = (x2 * cost + y2 * sint) / (BIGPROMOTE / PROMOTE);
            y1 = (y2 * cost - x2 * sint) / (BIGPROMOTE / PROMOTE);

            tmp->x = cx * PROMOTE + x1 / PROMOTE;
            tmp->y = cy * PROMOTE + y1 / PROMOTE;

            if(tmp->x < 0) tmp->x = 0;
            if(tmp->x >= PROMOTE * src->w) tmp->x = PROMOTE * src->w - 1;
            if(tmp->y < 0) tmp->y = 0;
            if(tmp->y >= PROMOTE * src->h) tmp->y = PROMOTE * src->h - 1;

            tmp++;
        }
    }

    /* 2: on fait le calcul global */
    for(j = 0; j < dy; j++)
    {
        for(i = 0; i < dx; i++)
        {
            /* On récupère les bornes du carré courant */
            uint32_t *p = dpix + j * RESOL * dstpitch + i * RESOL;
            int xA, yA, xT, yT, xU, yU, xV, yV;

            tmp = pt + (dx + 1) * j + i;
            tmp2 = tmp + dx + 1;

            xA = tmp[0].x * RESOL * RESOL;
            yA = tmp[0].y * RESOL * RESOL;

            xT = (tmp2[0].x - tmp[0].x) * RESOL;
            yT = (tmp2[0].y - tmp[0].y) * RESOL;

            xU = (tmp[1].x - tmp[0].x) * RESOL;
            yU = (tmp[1].y - tmp[0].y) * RESOL;

            xV = tmp[0].x + tmp2[1].x - tmp[1].x - tmp2[0].x;
            yV = tmp[0].y + tmp2[1].y - tmp[1].y - tmp2[0].y;

            /* Notre boucle critique interpole le carré courant */
            for(dj = 0; dj < RESOL; dj++)
            {
                int x = xA, y = yA;

                for(di = 0; di < RESOL; di++)
                {
                    *p++ = spix[y / (RESOL * RESOL * PROMOTE) * srcpitch
                                 + x / (RESOL * RESOL * PROMOTE)];
                    x += xU; y += yU;
                }

                xA += xT; yA += yT;
                xU += xV; yU += yV;

                p += dstpitch - RESOL;
            }
        }
    }

    return 0;
}

int main(int argc, char *argv[])
{
    SDL_Surface *tmp, *src, *dst;
    double phi, theta, gamma, beta, alpha;
    int cx, cy, R, d;
    uint8_t *p1, *p2;
    int action = 0;
    int n;

    cx = 612;
    cy = 481;
    R = 442;
    d = 400;
    phi = theta = gamma = beta = alpha = 0.0;

    if (-1 == SDL_Init(SDL_INIT_VIDEO))
    {
        printf("Can't initialize SDL");
        exit(1);
    }

    int options = (SDL_ANYFORMAT | SDL_SWSURFACE);
    SDL_Surface *screen = NULL;

    screen = SDL_SetVideoMode(640, 480, 0, options);
    if (NULL == screen)
    {
        printf("Can't set video mode");
        exit(1);
    }

    tmp = IMG_Load("input.png");
    src = SDL_CreateRGBSurface(SDL_SWSURFACE, tmp->w, tmp->h, 32,
                               0xff0000, 0xff00, 0xff, 0x0);
    SDL_BlitSurface(tmp, NULL, src, NULL);

    int done = 0;
    SDL_Event event;
    while (!done)
    {
        while (!done && SDL_PollEvent(&event))
        {
            switch(event.type)
            {
            case SDL_QUIT:
                done = 1;
                break;
            case SDL_KEYUP:
                action = 0;
                break;
            case SDL_KEYDOWN:
                action = event.key.keysym.sym;
                break;
            }
        }

        switch(action)
        {
        case SDLK_UP:
            phi += 0.02;
            if(phi > M_PI / 2) phi = M_PI / 2;
            beta -= 0.02;
            if(beta < -1.2) beta = -1.2;
            break;
        case SDLK_DOWN:
            phi -= 0.02;
            if(phi < -M_PI / 2) phi = -M_PI / 2;
            beta += 0.02;
            if(beta > 1.2) beta = 1.2;
            break;
        case SDLK_LEFT:
            gamma -= 0.02;
            alpha -= 0.02;
            if(alpha < -1.14) alpha = -1.14;
            break;
        case SDLK_RIGHT:
            gamma += 0.02;
            alpha += 0.02;
            if(alpha > 1.14) alpha = 1.14;
            break;
        case SDLK_PAGEUP:
            theta -= 0.02;
            break;
        case SDLK_PAGEDOWN:
            theta += 0.02;
            break;
        case '[':
            d *= 1.02;
            if(d > 900) d = 900;
            break;
        case ']':
            d *= 0.98;
            if(d < 90) d = 90;
            break;
        case SDLK_ESCAPE:
            done = 1;
            break;
        default:
            break;
        }

        /* si mode lat/lon */
        if(action)
        {
            phi = acos(cos(alpha) * cos(beta));
            theta = atan(sin(alpha) / tan(beta));
            if(beta < 0)
                theta += M_PI;
            gamma = -theta;
        }

        /* src, dst, centre X, centre Y, R, d, phi, theta, gamma */
        normal(src, screen, cx, cy, R, d, phi, theta, gamma);

        SDL_Flip(screen);
        SDL_Delay(10);
    }

    SDL_Quit();

    return 0;
}

